/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *   Mupen64plus - mempak.h                                                *
 *   Mupen64Plus homepage: https://mupen64plus.org/                        *
 *   Copyright (C) 2014 Bobby Smiles                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef M64P_DEVICE_SI_MEMPAK_H
#define M64P_DEVICE_SI_MEMPAK_H

#include <stddef.h>
#include <stdint.h>

struct storage_backend_interface;

#define DEFAULT_MEMPAK_DEVICEID UINT16_C(0x0001)
#define DEFAULT_MEMPAK_BANKS    UINT8_C(0x01)
#define DEFAULT_MEMPAK_VERSION  UINT8_C(0x00)

struct mempak
{
    void* storage;
    const struct storage_backend_interface* istorage;
};

enum { MEMPAK_SIZE = 0x8000 };

void format_mempak(uint8_t* mem,
    const uint32_t serial[6],
    uint16_t device_id,
    uint8_t banks,
    uint8_t version);

void init_mempak(struct mempak* mpk,
                 void* storage,
                 const struct storage_backend_interface* istorage);

extern const struct pak_interface g_imempak;

#endif
